# Copyright (c) 2015 by Tegile Systems, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Volume driver for Tegile storage.
"""

import ast
import json
import requests

from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import units
import six

from cinder import exception
from cinder import utils
from cinder.i18n import _, _LI, _LW
from cinder.volume import driver
from cinder.volume.drivers.san import san
from cinder.volume import utils as volume_utils
from cinder.zonemanager import utils as fczm_utils

LOG = logging.getLogger(__name__)
default_api_service = 'openstack'
TEGILE_API_PATH = 'zebi/api'
TEGILE_DEFAULT_BLOCK_SIZE = '32KB'
TEGILE_LOCAL_CONTAINER_NAME = 'Local'
DEBUG_LOGGING = False

tegile_opts = [
    cfg.StrOpt('tegile_default_pool',
               help='Create volumes in this pool'),
    cfg.StrOpt('tegile_default_project',
               help='Create volumes in this project')]

CONF = cfg.CONF
CONF.register_opts(tegile_opts)


def debugger(func):
    """Returns a wrapper that wraps func.

    The wrapper will log the entry and exit points of the function
    """

    def wrapper(*args, **kwds):
        if DEBUG_LOGGING:
            LOG.debug('Entering %(classname)s.%(funcname)s',
                      {'classname': args[0].__class__.__name__,
                       'funcname': func.__name__})
            LOG.debug('Arguments: %(args)s, %(kwds)s',
                      {'args': args[1:],
                       'kwds': kwds})
        f_result = func(*args, **kwds)
        if DEBUG_LOGGING:
            LOG.debug('Exiting %(classname)s.%(funcname)s',
                      {'classname': args[0].__class__.__name__,
                       'funcname': func.__name__})
            LOG.debug('Results: %(result)s',
                      {'result': f_result})
        return f_result

    return wrapper


class TegileAPIExecutor(object):
    def __init__(self, classname, hostname, username, password):
        self._classname = classname
        self._hostname = hostname
        self._username = username
        self._password = password

    @debugger
    @utils.retry(exceptions=(requests.ConnectionError, requests.Timeout))
    def send_api_request(self, method, params=None,
                         request_type='post',
                         api_service=default_api_service,
                         fine_logging=DEBUG_LOGGING):
        if params is not None:
            params = json.dumps(params)

        url = 'https://%s/%s/%s/%s' % (self._hostname,
                                       TEGILE_API_PATH,
                                       api_service,
                                       method)
        if fine_logging:
            LOG.debug('TegileAPIExecutor(%(classname)s) method: %(method)s, '
                      'url: %(url)s', {'classname': self._classname,
                                       'method': method,
                                       'url': url})
        if request_type == 'post':
            if fine_logging:
                LOG.debug('TegileAPIExecutor(%(classname)s) '
                          'method: %(method)s, payload: %(payload)s',
                          {'classname': self._classname,
                           'method': method,
                           'payload': params})
            req = requests.post(url,
                                data=params,
                                auth=(self._username, self._password),
                                verify=False)
        else:
            req = requests.get(url,
                               auth=(self._username, self._password),
                               verify=False)

        if fine_logging:
            LOG.debug('TegileAPIExecutor(%(classname)s) method: %(method)s, '
                      'return code: %(retcode)s',
                      {'classname': self._classname,
                       'method': method,
                       'retcode': req})
        try:
            response = req.json()
            if fine_logging:
                LOG.debug('TegileAPIExecutor(%(classname)s) '
                          'method: %(method)s, response: %(response)s',
                          {'classname': self._classname,
                           'method': method,
                           'response': response})
        except ValueError:
            response = ''
        req.close()

        if req.status_code != 200:
            msg = _('API response: %(response)s') % {'response': response}
            raise exception.TegileAPIException(msg)

        return response


class TegileIntelliFlashVolumeDriver(san.SanDriver):
    """Tegile IntelliFlash Volume Driver."""

    VENDOR = 'Tegile Systems Inc.'
    VERSION = '1.0.0'
    REQUIRED_OPTIONS = ['san_ip', 'san_login',
                        'san_password', 'tegile_default_pool']
    SNAPSHOT_PREFIX = 'Manual-V-'

    _api_executor = None

    def __init__(self, *args, **kwargs):
        self._context = None
        super(TegileIntelliFlashVolumeDriver, self).__init__(*args, **kwargs)
        self.configuration.append_config_values(tegile_opts)
        self._protocol = 'iSCSI'  # defaults to iscsi
        hostname = getattr(self.configuration, 'san_ip')
        username = getattr(self.configuration, 'san_login')
        password = getattr(self.configuration, 'san_password')
        self._default_pool = getattr(self.configuration, 'tegile_default_pool')
        self._default_project = (
            getattr(self.configuration, 'tegile_default_project') or
            'openstack')
        self._api_executor = TegileAPIExecutor(self.__class__.__name__,
                                               hostname,
                                               username,
                                               password)

    @debugger
    def do_setup(self, context):
        super(TegileIntelliFlashVolumeDriver, self).do_setup(context)
        self._context = context
        self._check_ops(self.REQUIRED_OPTIONS, self.configuration)

    @debugger
    def create_volume(self, volume):
        pool = volume_utils.extract_host(volume['host'], level='pool',
                                         default_pool_name=self._default_pool)
        tegile_volume = {'blockSize': TEGILE_DEFAULT_BLOCK_SIZE,
                         'datasetPath': '%s/%s/%s' %
                                        (pool,
                                         TEGILE_LOCAL_CONTAINER_NAME,
                                         self._default_project),
                         'local': 'true',
                         'name': volume['name'],
                         'poolName': '%s' % pool,
                         'projectName': '%s' % self._default_project,
                         'protocol': self._protocol,
                         'thinProvision': 'true',
                         'volSize': volume['size'] * units.Gi}
        params = list()
        params.append(tegile_volume)
        params.append(True)

        self._api_executor.send_api_request(method='createVolume',
                                            params=params)

        LOG.info(_LI("Created volume %(volname)s, volume id %(volid)s."),
                 {'volname': volume['name'], 'volid': volume['id']})

        return self.get_additional_info(volume, pool, self._default_project)

    @debugger
    def delete_volume(self, volume):
        """Deletes a snapshot."""
        params = list()
        pool, project, volume_name = self._get_pool_project_volume_name(volume)
        params.append('%s/%s/%s/%s' % (pool,
                                       TEGILE_LOCAL_CONTAINER_NAME,
                                       project,
                                       volume_name))
        params.append(True)
        params.append(False)

        self._api_executor.send_api_request('deleteVolume', params)

    @debugger
    def create_snapshot(self, snapshot):
        """Creates a snapshot."""
        snap_name = snapshot['name']
        display_list = [getattr(snapshot, 'display_name', ''),
                        getattr(snapshot, 'display_description', '')]
        snap_description = ':'.join(filter(None, display_list))
        # Limit to 254 characters
        snap_description = snap_description[:254]

        pool, project, volume_name = self._get_pool_project_volume_name(
            snapshot['volume'])

        volume = {'blockSize': TEGILE_DEFAULT_BLOCK_SIZE,
                  'datasetPath': '%s/%s/%s' %
                                 (pool,
                                  TEGILE_LOCAL_CONTAINER_NAME,
                                  project),
                  'local': 'true',
                  'name': volume_name,
                  'poolName': '%s' % pool,
                  'projectName': '%s' % project,
                  'protocol': self._protocol,
                  'thinProvision': 'true',
                  'volSize': snapshot['volume']['size'] * units.Gi}
        params = list()
        params.append(volume)
        params.append(snap_name)
        params.append(False)

        LOG.info(_LI('Creating snapshot for volume_name=%(vol)s'
                     ' snap_name=%(name)s snap_description=%(desc)s'),
                 {'vol': volume_name,
                  'name': snap_name,
                  'desc': snap_description})

        self._api_executor.send_api_request('createVolumeSnapshot', params)

    @debugger
    def delete_snapshot(self, snapshot):
        """Deletes a snapshot."""
        params = list()
        pool, project, volume_name = self._get_pool_project_volume_name(
            snapshot['volume'])
        params.append('%s/%s/%s/%s@%s%s' % (pool,
                                            TEGILE_LOCAL_CONTAINER_NAME,
                                            project,
                                            volume_name,
                                            self.SNAPSHOT_PREFIX,
                                            snapshot['name']))
        params.append(False)

        self._api_executor.send_api_request('deleteVolumeSnapshot', params)

    @debugger
    def create_volume_from_snapshot(self, volume, snapshot):
        """Creates a volume from snapshot."""
        params = list()
        pool, project, volume_name = self._get_pool_project_volume_name(
            snapshot['volume'])

        params.append('%s/%s/%s/%s@%s%s' % (pool,
                                            TEGILE_LOCAL_CONTAINER_NAME,
                                            project,
                                            volume_name,
                                            self.SNAPSHOT_PREFIX,
                                            snapshot['name']))
        params.append(volume['name'])
        params.append(True)
        params.append(True)

        self._api_executor.send_api_request('cloneVolumeSnapshot', params)
        return self.get_additional_info(volume, pool, project)

    @debugger
    def create_cloned_volume(self, volume, src_vref):
        """Creates a clone of the specified volume."""
        pool, project, volume_name = self._get_pool_project_volume_name(
            src_vref)
        data_set_path = '%s/%s/%s' % (pool,
                                      TEGILE_LOCAL_CONTAINER_NAME,
                                      project)
        source_volume = {'blockSize': TEGILE_DEFAULT_BLOCK_SIZE,
                         'datasetPath': data_set_path,
                         'local': 'true',
                         'name': volume_name,
                         'poolName': '%s' % pool,
                         'projectName': '%s' % project,
                         'protocol': self._protocol,
                         'thinProvision': 'true',
                         'volSize': src_vref['size'] * units.Gi}

        dest_volume = {'blockSize': TEGILE_DEFAULT_BLOCK_SIZE,
                       'datasetPath': data_set_path,
                       # clone can reside only in the source project
                       'local': 'true',
                       'name': volume['name'],
                       'poolName': '%s' % pool,
                       'projectName': '%s' % project,
                       'protocol': self._protocol,
                       'thinProvision': 'true',
                       'volSize': volume['size'] * units.Gi}

        params = list()
        params.append(source_volume)
        params.append(dest_volume)

        self._api_executor.send_api_request(method='createClonedVolume',
                                            params=params)
        return self.get_additional_info(volume, pool, project)

    @debugger
    def get_volume_stats(self, refresh=False):
        """Get volume status.

        If 'refresh' is True, run update first.
        The name is a bit misleading as
        the majority of the data here is cluster
        data
        """
        if refresh:
            try:
                self._update_volume_stats()
            except Exception:
                pass

        return self._stats

    @debugger
    def _update_volume_stats(self):
        """Retrieves stats info from volume group."""

        try:
            data = self._api_executor.send_api_request(method='getArrayStats',
                                                       request_type='get',
                                                       fine_logging=False)
            # fixing values coming back here as String to float
            data['total_capacity_gb'] = float(data.get('total_capacity_gb', 0))
            data['free_capacity_gb'] = float(data.get('free_capacity_gb', 0))
            for pool in data.get('pools', []):
                pool['total_capacity_gb'] = float(
                    pool.get('total_capacity_gb', 0))
                pool['free_capacity_gb'] = float(
                    pool.get('free_capacity_gb', 0))
                pool['allocated_capacity_gb'] = float(
                    pool.get('allocated_capacity_gb', 0))

            data['volume_backend_name'] = getattr(self.configuration,
                                                  'volume_backend_name')
            data['vendor_name'] = self.VENDOR
            data['driver_version'] = self.VERSION
            data['storage_protocol'] = self._protocol

            self._stats = data
        except Exception as e:
            LOG.warning(_LW('TegileIntelliFlashVolumeDriver(%(clsname)s) '
                            '_update_volume_stats failed: %(error)s'),
                        {'clsname': self.__class__.__name__,
                         'error': e})

    @debugger
    def get_pool(self, volume):
        """Returns pool name where volume resides.

        :param volume: The volume hosted by the driver.
        :return: Name of the pool where given volume is hosted.
        """
        pool = volume_utils.extract_host(volume['host'], level='pool',
                                         default_pool_name=self._default_pool)
        return pool

    @debugger
    def extend_volume(self, volume, new_size):
        params = list()
        pool, project, volume_name = self._get_pool_project_volume_name(volume)
        params.append('%s/%s/%s/%s' % (pool,
                                       TEGILE_LOCAL_CONTAINER_NAME,
                                       project,
                                       volume_name))
        vol_size = six.text_type(new_size)
        params.append(vol_size)
        params.append('GB')
        self._api_executor.send_api_request(method='resizeVolume',
                                            params=params)

    @debugger
    def manage_existing(self, volume, existing_ref):
        volume['name_id'] = existing_ref['name']
        pool, project, volume_name = self._get_pool_project_volume_name(volume)
        additional_info = self.get_additional_info(volume, pool, project)
        additional_info['_name_id'] = existing_ref['name'],
        return additional_info

    @debugger
    def manage_existing_get_size(self, volume, existing_ref):
        params = list()
        pool, project, volume_name = self._get_pool_project_volume_name(volume)
        params.append('%s/%s/%s/%s' % (pool,
                                       TEGILE_LOCAL_CONTAINER_NAME,
                                       project,
                                       existing_ref['name']))
        volume_size = self._api_executor.send_api_request(
            method='getVolumeSizeinGB',
            params=params)

        return volume_size

    @debugger
    def _get_pool_project_volume_name(self, volume):
        pool = volume_utils.extract_host(volume['host'], level='pool',
                                         default_pool_name=self._default_pool)
        try:
            project = volume['metadata']['project']
        except (AttributeError, TypeError, KeyError):
            project = self._default_project

        if volume['_name_id'] is not None:
            volume_name = volume['_name_id']
        else:
            volume_name = volume['name']

        return pool, project, volume_name

    @debugger
    def get_additional_info(self, volume, pool, project):
        try:
            metadata = self._get_volume_metadata(volume)
        except Exception:
            metadata = dict()
        metadata['pool'] = pool
        metadata['project'] = project
        return {'metadata': metadata}

    @debugger
    def _get_volume_metadata(self, volume):
        volume_metadata = {}
        if 'volume_metadata' in volume:
            for metadata in volume['volume_metadata']:
                volume_metadata[metadata['key']] = metadata['value']
        if 'metadata' in volume:
            metadata = volume['metadata']
            for key in metadata:
                volume_metadata[key] = metadata[key]
        return volume_metadata

    @debugger
    def _check_ops(self, required_ops, configuration):
        """Ensures that the options we care about are set."""
        for attr in required_ops:
            if not getattr(configuration, attr, None):
                raise exception.InvalidInput(reason=_('%(attr)s is not '
                                                      'set.') % {'attr': attr})


class TegileISCSIDriver(TegileIntelliFlashVolumeDriver, san.SanISCSIDriver):
    """Tegile ISCSI Driver."""

    def __init__(self, *args, **kwargs):
        super(TegileISCSIDriver, self).__init__(*args, **kwargs)
        self._protocol = 'iSCSI'

    @debugger
    def do_setup(self, context):
        super(TegileISCSIDriver, self).do_setup(context)

    @debugger
    def initialize_connection(self, volume, connector):
        """Driver entry point to attach a volume to an instance."""

        if getattr(self.configuration, 'use_chap_auth', False):
                chap_username = getattr(self.configuration,
                                        'chap_username',
                                        '')
                chap_password = getattr(self.configuration,
                                        'chap_password',
                                        '')
        else:
            chap_username = ''
            chap_password = ''

        if volume['provider_location'] is None:
            params = list()
            pool, project, volume_name = (
                self._get_pool_project_volume_name(volume))
            params.append('%s/%s/%s/%s' % (pool,
                                           TEGILE_LOCAL_CONTAINER_NAME,
                                           project,
                                           volume_name))
            initiator_info = {
                'initiatorName': connector['initiator'],
                'chapUserName': chap_username,
                'chapSecret': chap_password
            }
            params.append(initiator_info)
            mapping_info = self._api_executor.send_api_request(
                method='getISCSIMappingForVolume',
                params=params)
            target_portal = mapping_info['target_portal']
            target_iqn = mapping_info['target_iqn']
            target_lun = mapping_info['target_lun']
        else:
            (target_portal, target_iqn, target_lun) = (
                volume['provider_location'].split())

        connection_data = dict()
        connection_data['target_portal'] = target_portal
        connection_data['target_iqn'] = target_iqn
        connection_data['target_lun'] = target_lun
        connection_data['target_discovered'] = False,
        connection_data['volume_id'] = volume['id'],
        connection_data['discard'] = False
        if getattr(self.configuration, 'use_chap_auth', False):
            connection_data['auth_method'] = 'CHAP'
            connection_data['auth_username'] = chap_username
            connection_data['auth_password'] = chap_password
        return {
            'driver_volume_type': 'iscsi',
            'data': connection_data
        }

    @debugger
    def terminate_connection(self, volume, connector, **kwargs):
        pass

    @debugger
    def create_export(self, context, volume, connector):
        """Driver entry point to get the export info for a new volume."""
        params = list()
        pool, project, volume_name = self._get_pool_project_volume_name(volume)
        params.append('%s/%s/%s/%s' % (pool,
                                       TEGILE_LOCAL_CONTAINER_NAME,
                                       project,
                                       volume_name))
        if getattr(self.configuration, 'use_chap_auth', False):
            chap_username = getattr(self.configuration, 'chap_username', '')
            chap_password = getattr(self.configuration, 'chap_password', '')
        else:
            chap_username = ''
            chap_password = ''

        initiator_info = {
            'initiatorName': connector['initiator'],
            'chapUserName': chap_username,
            'chapSecret': chap_password
        }
        params.append(initiator_info)
        mapping_info = self._api_executor.send_api_request(
            method='getISCSIMappingForVolume',
            params=params)
        target_portal = mapping_info['target_portal']
        target_iqn = mapping_info['target_iqn']
        target_lun = mapping_info['target_lun']

        provider_location = '%s %s %s' % (target_portal,
                                          target_iqn,
                                          target_lun)
        if getattr(self.configuration, 'use_chap_auth', False):
            provider_auth = ('CHAP %s %s' % (chap_username,
                                             chap_password))
        else:
            provider_auth = None
        return (
            {'provider_location': provider_location,
             'provider_auth': provider_auth})


class TegileFCDriver(TegileIntelliFlashVolumeDriver,
                     driver.FibreChannelDriver):
    """Tegile FC driver."""

    def __init__(self, *args, **kwargs):
        super(TegileFCDriver, self).__init__(*args, **kwargs)
        self._protocol = 'FC'

    @debugger
    def do_setup(self, context):
        super(TegileFCDriver, self).do_setup(context)

    @fczm_utils.AddFCZone
    @debugger
    def initialize_connection(self, volume, connector):
        """Initializes the connection and returns connection info."""

        params = list()
        pool, project, volume_name = self._get_pool_project_volume_name(volume)
        params.append('%s/%s/%s/%s' % (pool,
                                       TEGILE_LOCAL_CONTAINER_NAME,
                                       project,
                                       volume_name))
        wwpns = connector['wwpns']

        connectors = ','.join(wwpns)

        params.append(connectors)
        target_info = self._api_executor.send_api_request(
            method='getFCPortsForVolume',
            params=params)
        initiator_target_map = target_info['initiator_target_map']
        connection_data = {
            'driver_volume_type': 'fibre_channel',
            'data': {
                'encrypted': False,
                'target_discovered': False,
                'target_lun': target_info['target_lun'],
                'target_wwn': ast.literal_eval(target_info['target_wwn']),
                'initiator_target_map': ast.literal_eval(initiator_target_map)
            }
        }

        return connection_data

    @fczm_utils.RemoveFCZone
    @debugger
    def terminate_connection(self, volume, connector, force=False, **kwargs):

        params = list()
        pool, project, volume_name = self._get_pool_project_volume_name(volume)
        params.append('%s/%s/%s/%s' % (pool,
                                       TEGILE_LOCAL_CONTAINER_NAME,
                                       project,
                                       volume_name))
        wwpns = connector['wwpns']

        connectors = ','.join(wwpns)

        params.append(connectors)
        target_info = self._api_executor.send_api_request(
            method='getFCPortsForVolume',
            params=params)
        initiator_target_map = target_info['initiator_target_map']

        connection_data = {
            'data': {
                'target_wwn': ast.literal_eval(target_info['target_wwn']),
                'initiator_target_map': ast.literal_eval(initiator_target_map)
            }
        }

        return connection_data



# Copyright 2012, Intel, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Unit Tests for cinder.volume.rpcapi
"""
import copy

import mock
from oslo_config import cfg
from oslo_serialization import jsonutils

from cinder import context
from cinder import db
from cinder import exception
from cinder import objects
from cinder import test
from cinder.tests.unit import fake_backup
from cinder.tests.unit import fake_constants as fake
from cinder.tests.unit import fake_snapshot
from cinder.tests.unit import fake_volume
from cinder.tests.unit import utils as tests_utils
from cinder.volume import rpcapi as volume_rpcapi
from cinder.volume import utils


CONF = cfg.CONF


class VolumeRpcAPITestCase(test.TestCase):

    def setUp(self):
        super(VolumeRpcAPITestCase, self).setUp()
        self.context = context.get_admin_context()
        vol = {}
        vol['host'] = 'fake_host'
        vol['availability_zone'] = CONF.storage_availability_zone
        vol['status'] = "available"
        vol['attach_status'] = "detached"
        vol['metadata'] = {"test_key": "test_val"}
        vol['size'] = 1
        volume = db.volume_create(self.context, vol)

        kwargs = {
            'status': "creating",
            'progress': '0%',
            'display_name': 'fake_name',
            'display_description': 'fake_description'}
        snapshot = tests_utils.create_snapshot(self.context, vol['id'],
                                               **kwargs)

        source_group = tests_utils.create_consistencygroup(
            self.context,
            availability_zone=CONF.storage_availability_zone,
            volume_type='type1,type2',
            host='fakehost@fakedrv#fakepool')

        cgsnapshot = tests_utils.create_cgsnapshot(
            self.context,
            consistencygroup_id=source_group.id)

        group = tests_utils.create_consistencygroup(
            self.context,
            availability_zone=CONF.storage_availability_zone,
            volume_type='type1,type2',
            host='fakehost@fakedrv#fakepool',
            cgsnapshot_id=cgsnapshot.id)

        group2 = tests_utils.create_consistencygroup(
            self.context,
            availability_zone=CONF.storage_availability_zone,
            volume_type='type1,type2',
            host='fakehost@fakedrv#fakepool',
            source_cgid=source_group.id)

        group = objects.ConsistencyGroup.get_by_id(self.context, group.id)
        group2 = objects.ConsistencyGroup.get_by_id(self.context, group2.id)
        cgsnapshot = objects.CGSnapshot.get_by_id(self.context, cgsnapshot.id)
        self.fake_volume = jsonutils.to_primitive(volume)
        self.fake_volume_obj = fake_volume.fake_volume_obj(self.context, **vol)
        self.fake_volume_metadata = volume["volume_metadata"]
        self.fake_snapshot = snapshot
        self.fake_reservations = ["RESERVATION"]
        self.fake_cg = group
        self.fake_cg2 = group2
        self.fake_src_cg = jsonutils.to_primitive(source_group)
        self.fake_cgsnap = cgsnapshot
        self.fake_backup_obj = fake_backup.fake_backup_obj(self.context)

    def test_serialized_volume_has_id(self):
        self.assertIn('id', self.fake_volume)

    def _test_volume_api(self, method, rpc_method, **kwargs):
        ctxt = context.RequestContext('fake_user', 'fake_project')

        if 'rpcapi_class' in kwargs:
            rpcapi_class = kwargs['rpcapi_class']
            del kwargs['rpcapi_class']
        else:
            rpcapi_class = volume_rpcapi.VolumeAPI
        rpcapi = rpcapi_class()
        expected_retval = 'foo' if method == 'call' else None

        target = {
            "version": kwargs.pop('version', rpcapi.RPC_API_VERSION)
        }

        if 'request_spec' in kwargs:
            spec = jsonutils.to_primitive(kwargs['request_spec'])
            kwargs['request_spec'] = spec

        expected_msg = copy.deepcopy(kwargs)
        if 'volume' in expected_msg:
            volume = expected_msg['volume']
            # NOTE(thangp): copy.deepcopy() is making oslo_versionedobjects
            # think that 'metadata' was changed.
            if isinstance(volume, objects.Volume):
                volume.obj_reset_changes()
            del expected_msg['volume']
            expected_msg['volume_id'] = volume['id']
            expected_msg['volume'] = volume
        if 'snapshot' in expected_msg:
            snapshot = expected_msg['snapshot']
            del expected_msg['snapshot']
            expected_msg['snapshot_id'] = snapshot.id
            expected_msg['snapshot'] = snapshot
        if 'cgsnapshot' in expected_msg:
            cgsnapshot = expected_msg['cgsnapshot']
            if cgsnapshot:
                cgsnapshot.consistencygroup
                kwargs['cgsnapshot'].consistencygroup
        if 'backup' in expected_msg:
            backup = expected_msg['backup']
            del expected_msg['backup']
            expected_msg['backup_id'] = backup.id
            expected_msg['backup'] = backup

        if 'host' in expected_msg:
            del expected_msg['host']
        if 'dest_host' in expected_msg:
            dest_host = expected_msg['dest_host']
            dest_host_dict = {'host': dest_host.host,
                              'capabilities': dest_host.capabilities}
            del expected_msg['dest_host']
            expected_msg['host'] = dest_host_dict
        if 'new_volume' in expected_msg:
            volume = expected_msg['new_volume']
            expected_msg['new_volume_id'] = volume['id']

        if 'host' in kwargs:
            host = kwargs['host']
        elif 'group' in kwargs:
            host = kwargs['group']['host']
        elif 'volume' in kwargs:
            host = kwargs['volume']['host']
        elif 'snapshot' in kwargs:
            host = 'fake_host'
        elif 'cgsnapshot' in kwargs:
            host = kwargs['cgsnapshot'].consistencygroup.host

        target['server'] = utils.extract_host(host)
        target['topic'] = '%s.%s' % (CONF.volume_topic, host)

        self.fake_args = None
        self.fake_kwargs = None

        def _fake_prepare_method(*args, **kwds):
            for kwd in kwds:
                self.assertEqual(kwds[kwd], target[kwd])
            return rpcapi.client

        def _fake_rpc_method(*args, **kwargs):
            self.fake_args = args
            self.fake_kwargs = kwargs
            if expected_retval:
                return expected_retval

        self.stubs.Set(rpcapi.client, "prepare", _fake_prepare_method)
        self.stubs.Set(rpcapi.client, rpc_method, _fake_rpc_method)

        retval = getattr(rpcapi, method)(ctxt, **kwargs)

        self.assertEqual(expected_retval, retval)
        expected_args = [ctxt, method]

        for arg, expected_arg in zip(self.fake_args, expected_args):
            self.assertEqual(expected_arg, arg)

        for kwarg, value in self.fake_kwargs.items():
            if isinstance(value, objects.Snapshot):
                expected_snapshot = expected_msg[kwarg].obj_to_primitive()
                snapshot = value.obj_to_primitive()
                self.assertEqual(expected_snapshot, snapshot)
            elif isinstance(value, objects.ConsistencyGroup):
                expected_cg = expected_msg[kwarg].obj_to_primitive()
                cg = value.obj_to_primitive()
                self.assertEqual(expected_cg, cg)
            elif isinstance(value, objects.CGSnapshot):
                expected_cgsnapshot = expected_msg[kwarg].obj_to_primitive()
                cgsnapshot = value.obj_to_primitive()
                self.assertEqual(expected_cgsnapshot, cgsnapshot)
            elif isinstance(value, objects.Volume):
                expected_volume = expected_msg[kwarg].obj_to_primitive()
                volume = value.obj_to_primitive()
                self.assertEqual(expected_volume, volume)
            elif isinstance(value, objects.Backup):
                expected_backup = expected_msg[kwarg].obj_to_primitive()
                backup = value.obj_to_primitive()
                self.assertEqual(expected_backup, backup)
            else:
                self.assertEqual(expected_msg[kwarg], value)

    @mock.patch('oslo_messaging.RPCClient.can_send_version', return_value=True)
    def test_create_consistencygroup(self, mock_can_send_version):
        self._test_volume_api('create_consistencygroup', rpc_method='cast',
                              group=self.fake_cg, host='fake_host1',
                              version='2.0')

        mock_can_send_version.return_value = False
        self._test_volume_api('create_consistencygroup', rpc_method='cast',
                              group=self.fake_cg, host='fake_host1',
                              version='1.26')

    @mock.patch('oslo_messaging.RPCClient.can_send_version', return_value=True)
    def test_delete_consistencygroup(self, mock_can_send_version):
        self._test_volume_api('delete_consistencygroup', rpc_method='cast',
                              group=self.fake_cg, version='2.0')

        mock_can_send_version.return_value = False
        self._test_volume_api('delete_consistencygroup', rpc_method='cast',
                              group=self.fake_cg, version='1.26')

    @mock.patch('oslo_messaging.RPCClient.can_send_version', return_value=True)
    def test_update_consistencygroup(self, mock_can_send_version):
        self._test_volume_api('update_consistencygroup', rpc_method='cast',
                              group=self.fake_cg, add_volumes=['vol1'],
                              remove_volumes=['vol2'], version='2.0')

        mock_can_send_version.return_value = False
        self._test_volume_api('update_consistencygroup', rpc_method='cast',
                              group=self.fake_cg, add_volumes=['vol1'],
                              remove_volumes=['vol2'], version='1.26')

    @mock.patch('oslo_messaging.RPCClient.can_send_version', return_value=True)
    def test_create_cgsnapshot(self, mock_can_send_version):
        self._test_volume_api('create_cgsnapshot', rpc_method='cast',
                              cgsnapshot=self.fake_cgsnap, version='2.0')

        mock_can_send_version.return_value = False
        self._test_volume_api('create_cgsnapshot', rpc_method='cast',
                              cgsnapshot=self.fake_cgsnap, version='1.31')

    @mock.patch('oslo_messaging.RPCClient.can_send_version', return_value=True)
    def test_delete_cgsnapshot(self, mock_can_send_version):
        self._test_volume_api('delete_cgsnapshot', rpc_method='cast',
                              cgsnapshot=self.fake_cgsnap, version='2.0')

        mock_can_send_version.return_value = False
        self._test_volume_api('delete_cgsnapshot', rpc_method='cast',
                              cgsnapshot=self.fake_cgsnap, version='1.31')

    @mock.patch('oslo_messaging.RPCClient.can_send_version',
                return_value=True)
    def test_create_volume(self, can_send_version):
        self._test_volume_api('create_volume',
                              rpc_method='cast',
                              volume=self.fake_volume_obj,
                              host='fake_host1',
                              request_spec='fake_request_spec',
                              filter_properties='fake_properties',
                              allow_reschedule=True,
                              version='2.0')
        can_send_version.assert_called_once_with('2.0')

    @mock.patch('oslo_messaging.RPCClient.can_send_version',
                return_value=False)
    def test_create_volume_old(self, can_send_version):
        # Tests backwards compatibility with older clients
        self._test_volume_api('create_volume',
                              rpc_method='cast',
                              volume=self.fake_volume_obj,
                              host='fake_host1',
                              request_spec='fake_request_spec',
                              filter_properties='fake_properties',
                              allow_reschedule=True,
                              version='1.24')
        can_send_version.assert_has_calls([mock.call('2.0'),
                                           mock.call('1.32')])

    @mock.patch('oslo_messaging.RPCClient.can_send_version',
                return_value=True)
    def test_create_volume_serialization(self, can_send_version):
        request_spec = {"metadata": self.fake_volume_metadata}
        self._test_volume_api('create_volume',
                              rpc_method='cast',
                              volume=self.fake_volume_obj,
                              host='fake_host1',
                              request_spec=request_spec,
                              filter_properties='fake_properties',
                              allow_reschedule=True,
                              version='2.0')
        can_send_version.assert_called_once_with('2.0')

    @mock.patch('oslo_messaging.RPCClient.can_send_version',
                return_value=True)
    def test_delete_volume(self, can_send_version):
        self._test_volume_api('delete_volume',
                              rpc_method='cast',
                              volume=self.fake_volume_obj,
                              unmanage_only=False,
                              cascade=False,
                              version='2.0')
        can_send_version.assert_called_once_with('2.0')

    @mock.patch('oslo_messaging.RPCClient.can_send_version',
                return_value=False)
    def test_delete_volume_old(self, can_send_version):
        self._test_volume_api('delete_volume',
                              rpc_method='cast',
                              volume=self.fake_volume_obj,
                              unmanage_only=False,
                              version='1.15')
        can_send_version.assert_has_calls([mock.call('2.0'),
                                           mock.call('1.40'),
                                           mock.call('1.33')])

    @mock.patch('oslo_messaging.RPCClient.can_send_version',
                return_value=True)
    def test_delete_volume_cascade(self, can_send_version):
        self._test_volume_api('delete_volume',
                              rpc_method='cast',
                              volume=self.fake_volume_obj,
                              unmanage_only=False,
                              cascade=True,
                              version='2.0')
        can_send_version.assert_any_call('2.0')

    @mock.patch('oslo_messaging.RPCClient.can_send_version', return_value=True)
    def test_create_snapshot(self, mock_can_send_version):
        self._test_volume_api('create_snapshot',
                              rpc_method='cast',
                              volume=self.fake_volume,
                              snapshot=self.fake_snapshot,
                              version='2.0')

        mock_can_send_version.return_value = False
        self._test_volume_api('create_snapshot',
                              rpc_method='cast',
                              volume=self.fake_volume,
                              snapshot=self.fake_snapshot,
                              version='1.20')

    @mock.patch('oslo_messaging.RPCClient.can_send_version', return_value=True)
    def test_delete_snapshot(self, mock_can_send_version):
        self._test_volume_api('delete_snapshot',
                              rpc_method='cast',
                              snapshot=self.fake_snapshot,
                              host='fake_host',
                              unmanage_only=False,
                              version='2.0')

        mock_can_send_version.return_value = False
        self._test_volume_api('delete_snapshot',
                              rpc_method='cast',
                              snapshot=self.fake_snapshot,
                              host='fake_host',
                              unmanage_only=False,
                              version='1.20')

    @mock.patch('oslo_messaging.RPCClient.can_send_version', return_value=True)
    def test_delete_snapshot_with_unmanage_only(self, mock_can_send_version):
        self._test_volume_api('delete_snapshot',
                              rpc_method='cast',
                              snapshot=self.fake_snapshot,
                              host='fake_host',
                              unmanage_only=True,
                              version='2.0')

        mock_can_send_version.return_value = False
        self._test_volume_api('delete_snapshot',
                              rpc_method='cast',
                              snapshot=self.fake_snapshot,
                              host='fake_host',
                              unmanage_only=True,
                              version='1.20')

    @mock.patch('oslo_messaging.RPCClient.can_send_version', return_value=True)
    def test_attach_volume_to_instance(self, mock_can_send_version):
        self._test_volume_api('attach_volume',
                              rpc_method='call',
                              volume=self.fake_volume,
                              instance_uuid='fake_uuid',
                              host_name=None,
                              mountpoint='fake_mountpoint',
                              mode='ro',
                              version='2.0')

        mock_can_send_version.return_value = False
        self._test_volume_api('attach_volume',
                              rpc_method='call',
                              volume=self.fake_volume,
                              instance_uuid='fake_uuid',
                              host_name=None,
                              mountpoint='fake_mountpoint',
                              mode='ro',
                              version='1.11')

    @mock.patch('oslo_messaging.RPCClient.can_send_version', return_value=True)
    def test_attach_volume_to_host(self, mock_can_send_version):
        self._test_volume_api('attach_volume',
                              rpc_method='call',
                              volume=self.fake_volume,
                              instance_uuid=None,
                              host_name='fake_host',
                              mountpoint='fake_mountpoint',
                              mode='rw',
                              version='2.0')

        mock_can_send_version.return_value = False
        self._test_volume_api('attach_volume',
                              rpc_method='call',
                              volume=self.fake_volume,
                              instance_uuid=None,
                              host_name='fake_host',
                              mountpoint='fake_mountpoint',
                              mode='rw',
                              version='1.11')

    @mock.patch('oslo_messaging.RPCClient.can_send_version', return_value=True)
    def test_detach_volume(self, mock_can_send_version):
        self._test_volume_api('detach_volume',
                              rpc_method='call',
                              volume=self.fake_volume,
                              attachment_id='fake_uuid',
                              version="2.0")

        mock_can_send_version.return_value = False
        self._test_volume_api('detach_volume',
                              rpc_method='call',
                              volume=self.fake_volume,
                              attachment_id='fake_uuid',
                              version="1.20")

    @mock.patch('oslo_messaging.RPCClient.can_send_version', return_value=True)
    def test_copy_volume_to_image(self, mock_can_send_version):
        self._test_volume_api('copy_volume_to_image',
                              rpc_method='cast',
                              volume=self.fake_volume,
                              image_meta={'id': 'fake_image_id',
                                          'container_format': 'fake_type',
                                          'disk_format': 'fake_type'},
                              version='2.0')

        mock_can_send_version.return_value = False
        self._test_volume_api('copy_volume_to_image',
                              rpc_method='cast',
                              volume=self.fake_volume,
                              image_meta={'id': 'fake_image_id',
                                          'container_format': 'fake_type',
                                          'disk_format': 'fake_type'},
                              version='1.3')

    @mock.patch('oslo_messaging.RPCClient.can_send_version', return_value=True)
    def test_initialize_connection(self, mock_can_send_version):
        self._test_volume_api('initialize_connection',
                              rpc_method='call',
                              volume=self.fake_volume,
                              connector='fake_connector',
                              version='2.0')

        mock_can_send_version.return_value = False
        self._test_volume_api('initialize_connection',
                              rpc_method='call',
                              volume=self.fake_volume,
                              connector='fake_connector',
                              version='1.0')

    @mock.patch('oslo_messaging.RPCClient.can_send_version', return_value=True)
    def test_terminate_connection(self, mock_can_send_version):
        self._test_volume_api('terminate_connection',
                              rpc_method='call',
                              volume=self.fake_volume,
                              connector='fake_connector',
                              force=False,
                              version='2.0')

        mock_can_send_version.return_value = False
        self._test_volume_api('terminate_connection',
                              rpc_method='call',
                              volume=self.fake_volume,
                              connector='fake_connector',
                              force=False,
                              version='1.0')

    @mock.patch('oslo_messaging.RPCClient.can_send_version', return_value=True)
    def test_accept_transfer(self, mock_can_send_version):
        self._test_volume_api('accept_transfer',
                              rpc_method='call',
                              volume=self.fake_volume,
                              new_user='e5565fd0-06c8-11e3-'
                                       '8ffd-0800200c9b77',
                              new_project='e4465fd0-06c8-11e3'
                                          '-8ffd-0800200c9a66',
                              version='2.0')

        mock_can_send_version.return_value = False
        self._test_volume_api('accept_transfer',
                              rpc_method='call',
                              volume=self.fake_volume,
                              new_user='e5565fd0-06c8-11e3-'
                                       '8ffd-0800200c9b77',
                              new_project='e4465fd0-06c8-11e3'
                                          '-8ffd-0800200c9a66',
                              version='1.9')

    @mock.patch('oslo_messaging.RPCClient.can_send_version',
                return_value=True)
    def test_extend_volume(self, can_send_version):
        self._test_volume_api('extend_volume',
                              rpc_method='cast',
                              volume=self.fake_volume_obj,
                              new_size=1,
                              reservations=self.fake_reservations,
                              version='2.0')
        can_send_version.assert_called_once_with('2.0')

    @mock.patch('oslo_messaging.RPCClient.can_send_version',
                return_value=False)
    def test_extend_volume_old(self, can_send_version):
        self._test_volume_api('extend_volume',
                              rpc_method='cast',
                              volume=self.fake_volume_obj,
                              new_size=1,
                              reservations=self.fake_reservations,
                              version='1.14')
        can_send_version.assert_has_calls([mock.call('2.0'),
                                           mock.call('1.35')])

    @mock.patch('oslo_messaging.RPCClient.can_send_version',
                return_value=True)
    def test_migrate_volume(self, can_send_version):
        class FakeHost(object):
            def __init__(self):
                self.host = 'host'
                self.capabilities = {}
        dest_host = FakeHost()
        self._test_volume_api('migrate_volume',
                              rpc_method='cast',
                              volume=self.fake_volume_obj,
                              dest_host=dest_host,
                              force_host_copy=True,
                              version='2.0')
        can_send_version.assert_called_once_with('2.0')

    @mock.patch('oslo_messaging.RPCClient.can_send_version',
                return_value=False)
    def test_migrate_volume_old(self, can_send_version):
        class FakeHost(object):
            def __init__(self):
                self.host = 'host'
                self.capabilities = {}
        dest_host = FakeHost()
        self._test_volume_api('migrate_volume',
                              rpc_method='cast',
                              volume=self.fake_volume_obj,
                              dest_host=dest_host,
                              force_host_copy=True,
                              version='1.8')
        can_send_version.assert_has_calls([mock.call('2.0'),
                                           mock.call('1.36')])

    @mock.patch('oslo_messaging.RPCClient.can_send_version',
                return_value=True)
    def test_migrate_volume_completion(self, can_send_version):
        self._test_volume_api('migrate_volume_completion',
                              rpc_method='call',
                              volume=self.fake_volume_obj,
                              new_volume=self.fake_volume_obj,
                              error=False,
                              version='2.0')
        can_send_version.assert_called_once_with('2.0')

    @mock.patch('oslo_messaging.RPCClient.can_send_version',
                return_value=False)
    def test_migrate_volume_completion_old(self, can_send_version):
        self._test_volume_api('migrate_volume_completion',
                              rpc_method='call',
                              volume=self.fake_volume_obj,
                              new_volume=self.fake_volume_obj,
                              error=False,
                              version='1.10')
        can_send_version.assert_has_calls([mock.call('2.0'),
                                           mock.call('1.36')])

    @mock.patch('oslo_messaging.RPCClient.can_send_version', return_value=True)
    @mock.patch('cinder.quota.DbQuotaDriver.rollback')
    def test_retype(self, rollback, can_send_version):
        class FakeHost(object):
            def __init__(self):
                self.host = 'host'
                self.capabilities = {}
        dest_host = FakeHost()
        self._test_volume_api('retype',
                              rpc_method='cast',
                              volume=self.fake_volume_obj,
                              new_type_id='fake',
                              dest_host=dest_host,
                              migration_policy='never',
                              reservations=self.fake_reservations,
                              old_reservations=self.fake_reservations,
                              version='2.0')
        rollback.assert_not_called()
        can_send_version.assert_called_once_with('2.0')

    @mock.patch('oslo_messaging.RPCClient.can_send_version',
                side_effect=[False, True])
    @mock.patch('cinder.quota.DbQuotaDriver.rollback')
    def test_retype_137(self, rollback, can_send_version):
        class FakeHost(object):
            def __init__(self):
                self.host = 'host'
                self.capabilities = {}
        dest_host = FakeHost()
        self._test_volume_api('retype',
                              rpc_method='cast',
                              volume=self.fake_volume_obj,
                              new_type_id='fake',
                              dest_host=dest_host,
                              migration_policy='never',
                              reservations=self.fake_reservations,
                              old_reservations=self.fake_reservations,
                              version='1.37')
        rollback.assert_not_called()
        can_send_version.assert_any_call('2.0')
        can_send_version.assert_any_call('1.37')

    @mock.patch('cinder.quota.DbQuotaDriver.rollback')
    @mock.patch('oslo_messaging.RPCClient.can_send_version',
                side_effect=[False, False, True])
    def test_retype_version_134(self, can_send_version, rollback):
        class FakeHost(object):
            def __init__(self):
                self.host = 'host'
                self.capabilities = {}
        dest_host = FakeHost()
        self._test_volume_api('retype',
                              rpc_method='cast',
                              volume=self.fake_volume_obj,
                              new_type_id='fake',
                              dest_host=dest_host,
                              migration_policy='never',
                              reservations=self.fake_reservations,
                              old_reservations=self.fake_reservations,
                              version='1.34')
        self.assertTrue(rollback.called)
        can_send_version.assert_any_call('2.0')
        can_send_version.assert_any_call('1.37')
        can_send_version.assert_any_call('1.34')

    @mock.patch('cinder.quota.DbQuotaDriver.rollback')
    @mock.patch('oslo_messaging.RPCClient.can_send_version',
                side_effect=[False, False, False])
    def test_retype_version_112(self, can_send_version, rollback):
        class FakeHost(object):
            def __init__(self):
                self.host = 'host'
                self.capabilities = {}
        dest_host = FakeHost()
        self._test_volume_api('retype',
                              rpc_method='cast',
                              volume=self.fake_volume_obj,
                              new_type_id='fake',
                              dest_host=dest_host,
                              migration_policy='never',
                              reservations=self.fake_reservations,
                              old_reservations=self.fake_reservations,
                              version='1.12')
        self.assertTrue(rollback.called)
        can_send_version.assert_any_call('1.37')
        can_send_version.assert_any_call('1.34')

    @mock.patch('oslo_messaging.RPCClient.can_send_version', return_value=True)
    def test_manage_existing(self, mock_can_send_version):
        self._test_volume_api('manage_existing',
                              rpc_method='cast',
                              volume=self.fake_volume,
                              ref={'lv_name': 'foo'},
                              version='2.0')

        mock_can_send_version.return_value = False
        self._test_volume_api('manage_existing',
                              rpc_method='cast',
                              volume=self.fake_volume,
                              ref={'lv_name': 'foo'},
                              version='1.15')

    @mock.patch('oslo_messaging.RPCClient.can_send_version', return_value=True)
    def test_manage_existing_snapshot(self, mock_can_send_version):
        volume_update = {'host': 'fake_host'}
        snpshot = {
            'id': fake.snapshot_id,
            'volume_id': fake.volume_id,
            'status': "creating",
            'progress': '0%',
            'volume_size': 0,
            'display_name': 'fake_name',
            'display_description': 'fake_description',
            'volume': fake_volume.fake_db_volume(**volume_update),
            'expected_attrs': ['volume'], }
        my_fake_snapshot_obj = fake_snapshot.fake_snapshot_obj(self.context,
                                                               **snpshot)
        self._test_volume_api('manage_existing_snapshot',
                              rpc_method='cast',
                              snapshot=my_fake_snapshot_obj,
                              ref='foo',
                              host='fake_host',
                              version='2.0')

        mock_can_send_version.return_value = False
        self._test_volume_api('manage_existing_snapshot',
                              rpc_method='cast',
                              snapshot=my_fake_snapshot_obj,
                              ref='foo',
                              host='fake_host',
                              version='1.28')

    @mock.patch('oslo_messaging.RPCClient.can_send_version', return_value=True)
    def test_promote_replica(self, mock_can_send_version):
        self._test_volume_api('promote_replica',
                              rpc_method='cast',
                              volume=self.fake_volume,
                              version='2.0')

        mock_can_send_version.return_value = False
        self._test_volume_api('promote_replica',
                              rpc_method='cast',
                              volume=self.fake_volume,
                              version='1.17')

    @mock.patch('oslo_messaging.RPCClient.can_send_version', return_value=True)
    def test_reenable_replica(self, mock_can_send_version):
        self._test_volume_api('reenable_replication',
                              rpc_method='cast',
                              volume=self.fake_volume,
                              version='2.0')

        mock_can_send_version.return_value = False
        self._test_volume_api('reenable_replication',
                              rpc_method='cast',
                              volume=self.fake_volume,
                              version='1.17')

    @mock.patch('oslo_messaging.RPCClient.can_send_version', return_value=True)
    def test_freeze_host(self, mock_can_send_version):
        self._test_volume_api('freeze_host', rpc_method='call',
                              host='fake_host', version='2.0')

        mock_can_send_version.return_value = False
        self._test_volume_api('freeze_host', rpc_method='call',
                              host='fake_host', version='1.39')

    @mock.patch('oslo_messaging.RPCClient.can_send_version', return_value=True)
    def test_thaw_host(self, mock_can_send_version):
        self._test_volume_api('thaw_host', rpc_method='call', host='fake_host',
                              version='2.0')

        mock_can_send_version.return_value = False
        self._test_volume_api('thaw_host', rpc_method='call', host='fake_host',
                              version='1.39')

    @mock.patch('oslo_messaging.RPCClient.can_send_version', return_value=True)
    def test_failover_host(self, mock_can_send_version):
        self._test_volume_api('failover_host', rpc_method='cast',
                              host='fake_host',
                              secondary_backend_id='fake_backend',
                              version='2.0')

        mock_can_send_version.return_value = False
        self._test_volume_api('failover_host', rpc_method='cast',
                              host='fake_host',
                              secondary_backend_id='fake_backend',
                              version='1.39')

    @mock.patch('oslo_messaging.RPCClient.can_send_version', return_value=True)
    def test_create_consistencygroup_from_src_cgsnapshot(
            self, mock_can_send_version):
        self._test_volume_api('create_consistencygroup_from_src',
                              rpc_method='cast',
                              group=self.fake_cg,
                              cgsnapshot=self.fake_cgsnap,
                              source_cg=None,
                              version='2.0')

        mock_can_send_version.return_value = False
        self._test_volume_api('create_consistencygroup_from_src',
                              rpc_method='cast',
                              group=self.fake_cg,
                              cgsnapshot=self.fake_cgsnap,
                              source_cg=None,
                              version='1.31')

    @mock.patch('oslo_messaging.RPCClient.can_send_version', return_value=True)
    def test_create_consistencygroup_from_src_cg(self, mock_can_send_version):
        self._test_volume_api('create_consistencygroup_from_src',
                              rpc_method='cast',
                              group=self.fake_cg2,
                              cgsnapshot=None,
                              source_cg=self.fake_src_cg,
                              version='2.0')

        mock_can_send_version.return_value = False
        self._test_volume_api('create_consistencygroup_from_src',
                              rpc_method='cast',
                              group=self.fake_cg2,
                              cgsnapshot=None,
                              source_cg=self.fake_src_cg,
                              version='1.31')

    @mock.patch('oslo_messaging.RPCClient.can_send_version', return_value=True)
    def test_get_capabilities(self, mock_can_send_version):
        self._test_volume_api('get_capabilities',
                              rpc_method='call',
                              host='fake_host',
                              discover=True,
                              version='2.0')

        mock_can_send_version.return_value = False
        self._test_volume_api('get_capabilities',
                              rpc_method='call',
                              host='fake_host',
                              discover=True,
                              version='1.29')

    @mock.patch('oslo_messaging.RPCClient.can_send_version', return_value=True)
    def test_remove_export(self, mock_can_send_version):
        self._test_volume_api('remove_export',
                              rpc_method='cast',
                              volume=self.fake_volume,
                              version='2.0')

        mock_can_send_version.return_value = False
        self._test_volume_api('remove_export',
                              rpc_method='cast',
                              volume=self.fake_volume,
                              version='1.30')

    @mock.patch('oslo_messaging.RPCClient.can_send_version', return_value=True)
    def test_get_backup_device(self, mock_can_send_version):
        self._test_volume_api('get_backup_device',
                              rpc_method='call',
                              backup=self.fake_backup_obj,
                              volume=self.fake_volume_obj,
                              version='2.0')

        mock_can_send_version.return_value = False
        self.assertRaises(exception.ServiceTooOld, self._test_volume_api,
                          'get_backup_device', rpc_method='call',
                          backup=self.fake_backup_obj,
                          volume=self.fake_volume_obj, version='1.38')

    @mock.patch('oslo_messaging.RPCClient.can_send_version', return_value=True)
    def test_secure_file_operations_enabled(self, mock_can_send_version):
        self._test_volume_api('secure_file_operations_enabled',
                              rpc_method='call',
                              volume=self.fake_volume_obj,
                              version='2.0')

        mock_can_send_version.return_value = False
        self.assertRaises(exception.ServiceTooOld, self._test_volume_api,
                          'secure_file_operations_enabled', rpc_method='call',
                          volume=self.fake_volume_obj, version='1.38')



# Copyright (c) 2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime

from iso8601 import iso8601
from lxml import etree
from oslo_utils import timeutils
import webob.exc

from cinder.api.contrib import hosts as os_hosts
from cinder import context
from cinder import db
from cinder import test


created_time = datetime.datetime(2012, 11, 14, 1, 20, 41, 95099)
curr_time = datetime.datetime(2013, 7, 3, 0, 0, 1)

SERVICE_LIST = [
    {'created_at': created_time, 'updated_at': curr_time,
     'host': 'test.host.1', 'topic': 'cinder-volume', 'disabled': 0,
     'availability_zone': 'cinder'},
    {'created_at': created_time, 'updated_at': curr_time,
     'host': 'test.host.1', 'topic': 'cinder-volume', 'disabled': 0,
     'availability_zone': 'cinder'},
    {'created_at': created_time, 'updated_at': curr_time,
     'host': 'test.host.1', 'topic': 'cinder-volume', 'disabled': 0,
     'availability_zone': 'cinder'},
    {'created_at': created_time, 'updated_at': curr_time,
     'host': 'test.host.1', 'topic': 'cinder-volume', 'disabled': 0,
     'availability_zone': 'cinder'},
    {'created_at': created_time, 'updated_at': None,
     'host': 'test.host.1', 'topic': 'cinder-volume', 'disabled': 0,
     'availability_zone': 'cinder'},
]

LIST_RESPONSE = [{'service-status': 'available', 'service': 'cinder-volume',
                  'zone': 'cinder', 'service-state': 'enabled',
                  'host_name': 'test.host.1', 'last-update': curr_time},
                 {'service-status': 'available', 'service': 'cinder-volume',
                  'zone': 'cinder', 'service-state': 'enabled',
                  'host_name': 'test.host.1', 'last-update': curr_time},
                 {'service-status': 'available', 'service': 'cinder-volume',
                  'zone': 'cinder', 'service-state': 'enabled',
                  'host_name': 'test.host.1', 'last-update': curr_time},
                 {'service-status': 'available', 'service': 'cinder-volume',
                  'zone': 'cinder', 'service-state': 'enabled',
                  'host_name': 'test.host.1', 'last-update': curr_time},
                 {'service-status': 'unavailable', 'service': 'cinder-volume',
                  'zone': 'cinder', 'service-state': 'enabled',
                  'host_name': 'test.host.1', 'last-update': None},
                 ]


def stub_utcnow(with_timezone=False):
    tzinfo = iso8601.Utc() if with_timezone else None
    return datetime.datetime(2013, 7, 3, 0, 0, 2, tzinfo=tzinfo)


def stub_service_get_all(context, filters=None):
    return SERVICE_LIST


class FakeRequest(object):
    environ = {'cinder.context': context.get_admin_context()}
    GET = {}


class FakeRequestWithcinderZone(object):
    environ = {'cinder.context': context.get_admin_context()}
    GET = {'zone': 'cinder'}


class HostTestCase(test.TestCase):
    """Test Case for hosts."""

    def setUp(self):
        super(HostTestCase, self).setUp()
        self.controller = os_hosts.HostController()
        self.req = FakeRequest()
        self.stubs.Set(db, 'service_get_all',
                       stub_service_get_all)
        self.stubs.Set(timeutils, 'utcnow', stub_utcnow)

    def _test_host_update(self, host, key, val, expected_value):
        body = {key: val}
        result = self.controller.update(self.req, host, body=body)
        self.assertEqual(expected_value, result[key])

    def test_list_hosts(self):
        """Verify that the volume hosts are returned."""
        hosts = os_hosts._list_hosts(self.req)
        self.assertEqual(LIST_RESPONSE, hosts)

        cinder_hosts = os_hosts._list_hosts(self.req, 'cinder-volume')
        expected = [host for host in LIST_RESPONSE
                    if host['service'] == 'cinder-volume']
        self.assertEqual(expected, cinder_hosts)

    def test_list_hosts_with_zone(self):
        req = FakeRequestWithcinderZone()
        hosts = os_hosts._list_hosts(req)
        self.assertEqual(LIST_RESPONSE, hosts)

    def test_bad_status_value(self):
        self.assertRaises(webob.exc.HTTPBadRequest, self.controller.update,
                          self.req, 'test.host.1', body={'status': 'bad'})
        self.assertRaises(webob.exc.HTTPBadRequest,
                          self.controller.update,
                          self.req,
                          'test.host.1',
                          body={'status': 'disablabc'})

    def test_bad_update_key(self):
        bad_body = {'crazy': 'bad'}
        self.assertRaises(webob.exc.HTTPBadRequest, self.controller.update,
                          self.req, 'test.host.1', body=bad_body)

    def test_bad_update_key_and_correct_udpate_key(self):
        bad_body = {'status': 'disable', 'crazy': 'bad'}
        self.assertRaises(webob.exc.HTTPBadRequest, self.controller.update,
                          self.req, 'test.host.1', body=bad_body)

    def test_good_udpate_keys(self):
        body = {'status': 'disable'}
        self.assertRaises(NotImplementedError, self.controller.update,
                          self.req, 'test.host.1', body=body)

    def test_bad_host(self):
        self.assertRaises(webob.exc.HTTPNotFound,
                          self.controller.update,
                          self.req,
                          'bogus_host_name',
                          body={'disabled': 0})

    def test_show_forbidden(self):
        self.req.environ['cinder.context'].is_admin = False
        dest = 'dummydest'
        self.assertRaises(webob.exc.HTTPForbidden,
                          self.controller.show,
                          self.req, dest)
        self.req.environ['cinder.context'].is_admin = True

    def test_show_host_not_exist(self):
        """A host given as an argument does not exists."""
        self.req.environ['cinder.context'].is_admin = True
        dest = 'dummydest'
        self.assertRaises(webob.exc.HTTPNotFound,
                          self.controller.show,
                          self.req, dest)


class HostSerializerTest(test.TestCase):
    def setUp(self):
        super(HostSerializerTest, self).setUp()
        self.deserializer = os_hosts.HostDeserializer()

    def test_index_serializer(self):
        serializer = os_hosts.HostIndexTemplate()
        text = serializer.serialize({"hosts": LIST_RESPONSE})

        tree = etree.fromstring(text)

        self.assertEqual('hosts', tree.tag)
        self.assertEqual(len(LIST_RESPONSE), len(tree))
        for i in range(len(LIST_RESPONSE)):
            self.assertEqual('host', tree[i].tag)
            self.assertEqual(LIST_RESPONSE[i]['service-status'],
                             tree[i].get('service-status'))
            self.assertEqual(LIST_RESPONSE[i]['service'],
                             tree[i].get('service'))
            self.assertEqual(LIST_RESPONSE[i]['zone'],
                             tree[i].get('zone'))
            self.assertEqual(LIST_RESPONSE[i]['service-state'],
                             tree[i].get('service-state'))
            self.assertEqual(LIST_RESPONSE[i]['host_name'],
                             tree[i].get('host_name'))
            self.assertEqual(str(LIST_RESPONSE[i]['last-update']),
                             tree[i].get('last-update'))

    def test_update_serializer_with_status(self):
        exemplar = dict(host='test.host.1', status='enabled')
        serializer = os_hosts.HostUpdateTemplate()
        text = serializer.serialize(exemplar)

        tree = etree.fromstring(text)

        self.assertEqual('host', tree.tag)
        for key, value in exemplar.items():
            self.assertEqual(value, tree.get(key))

    def test_update_deserializer(self):
        exemplar = dict(status='enabled', foo='bar')
        intext = ("<?xml version='1.0' encoding='UTF-8'?>\n"
                  '<updates><status>enabled</status><foo>bar</foo></updates>')
        result = self.deserializer.deserialize(intext)

        self.assertEqual(dict(body=exemplar), result)



#    Copyright 2012 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_utils import timeutils

from cinder import exception
from cinder.objects import fields
from cinder.tests.unit.brick import fake_lvm
from cinder.volume import driver
from cinder.volume.drivers import lvm
from cinder.zonemanager import utils as fczm_utils


class FakeISCSIDriver(lvm.LVMVolumeDriver):
    """Logs calls instead of executing."""
    def __init__(self, *args, **kwargs):
        super(FakeISCSIDriver, self).__init__(execute=self.fake_execute,
                                              *args, **kwargs)
        self.vg = fake_lvm.FakeBrickLVM('cinder-volumes', False,
                                        None, 'default',
                                        self.fake_execute)

    def check_for_setup_error(self):
        """No setup necessary in fake mode."""
        pass

    def initialize_connection(self, volume, connector):
        volume_metadata = {}

        for metadata in volume['volume_admin_metadata']:
            volume_metadata[metadata['key']] = metadata['value']

        access_mode = volume_metadata.get('attached_mode')
        if access_mode is None:
            access_mode = ('ro'
                           if volume_metadata.get('readonly') == 'True'
                           else 'rw')

        return {'driver_volume_type': 'iscsi',
                'data': {'access_mode': access_mode}}

    def terminate_connection(self, volume, connector, **kwargs):
        pass

    @staticmethod
    def fake_execute(cmd, *_args, **_kwargs):
        """Execute that simply logs the command."""
        return (None, None)


class FakeISERDriver(FakeISCSIDriver):
    """Logs calls instead of executing."""
    def __init__(self, *args, **kwargs):
        super(FakeISERDriver, self).__init__(execute=self.fake_execute,
                                             *args, **kwargs)

    def initialize_connection(self, volume, connector):
        return {
            'driver_volume_type': 'iser',
            'data': {}
        }

    @staticmethod
    def fake_execute(cmd, *_args, **_kwargs):
        """Execute that simply logs the command."""
        return (None, None)


class FakeFibreChannelDriver(driver.FibreChannelDriver):

    @fczm_utils.AddFCZone
    def initialize_connection(self, volume, connector):
        return {
            'driver_volume_type': 'fibre_channel',
            'data': {
                'initiator_target_map': {'fake_wwn': ['fake_wwn2']},
            }}

    @fczm_utils.AddFCZone
    def no_zone_initialize_connection(self, volume, connector):
        """This shouldn't call the ZM."""
        return {
            'driver_volume_type': 'bogus',
            'data': {
                'initiator_target_map': {'fake_wwn': ['fake_wwn2']},
            }}

    @fczm_utils.RemoveFCZone
    def terminate_connection(self, volume, connector, **kwargs):
        return {
            'driver_volume_type': 'fibre_channel',
            'data': {
                'initiator_target_map': {'fake_wwn': ['fake_wwn2']},
            }}

    @fczm_utils.RemoveFCZone
    def no_zone_terminate_connection(self, volume, connector, **kwargs):
        return {
            'driver_volume_type': 'bogus',
            'data': {
                'initiator_target_map': {'fake_wwn': ['fake_wwn2']},
            }}


class LoggingVolumeDriver(driver.VolumeDriver):
    """Logs and records calls, for unit tests."""

    def check_for_setup_error(self):
        pass

    def create_volume(self, volume):
        self.log_action('create_volume', volume)

    def delete_volume(self, volume):
        self.clear_volume(volume)
        self.log_action('delete_volume', volume)

    def clear_volume(self, volume):
        self.log_action('clear_volume', volume)

    def local_path(self, volume):
        raise NotImplementedError()

    def ensure_export(self, context, volume):
        self.log_action('ensure_export', volume)

    def create_export(self, context, volume):
        self.log_action('create_export', volume)

    def remove_export(self, context, volume):
        self.log_action('remove_export', volume)

    def initialize_connection(self, volume, connector):
        self.log_action('initialize_connection', volume)

    def terminate_connection(self, volume, connector):
        self.log_action('terminate_connection', volume)

    def create_export_snapshot(self, context, snapshot):
        self.log_action('create_export_snapshot', snapshot)

    def remove_export_snapshot(self, context, snapshot):
        self.log_action('remove_export_snapshot', snapshot)

    def initialize_connection_snapshot(self, snapshot, connector):
        self.log_action('initialize_connection_snapshot', snapshot)

    def terminate_connection_snapshot(self, snapshot, connector):
        self.log_action('terminate_connection_snapshot', snapshot)

    def create_cloned_volume(self, volume, src_vol):
        self.log_action('create_cloned_volume', volume)

    _LOGS = []

    @staticmethod
    def clear_logs():
        LoggingVolumeDriver._LOGS = []

    @staticmethod
    def log_action(action, parameters):
        """Logs the command."""
        log_dictionary = {}
        if parameters:
            log_dictionary = dict(parameters)
        log_dictionary['action'] = action
        LoggingVolumeDriver._LOGS.append(log_dictionary)

    @staticmethod
    def all_logs():
        return LoggingVolumeDriver._LOGS

    @staticmethod
    def logs_like(action, **kwargs):
        matches = []
        for entry in LoggingVolumeDriver._LOGS:
            if entry['action'] != action:
                continue
            match = True
            for k, v in kwargs.items():
                if entry.get(k) != v:
                    match = False
                    break
            if match:
                matches.append(entry)
        return matches


class FakeGateDriver(lvm.LVMVolumeDriver):
    """Class designation for FakeGateDriver.

    FakeGateDriver is for TESTING ONLY. There are a few
    driver features such as CG and replication that are not
    supported by the reference driver LVM currently. Adding
    those functions in this fake driver will help detect
    problems when changes are introduced in those functions.

    Implementation of this driver is NOT meant for production.
    They are implemented simply to make sure calls to the driver
    functions are passing in the correct parameters, and the
    results returned by the driver are handled properly by the manager.

    """
    def __init__(self, *args, **kwargs):
        super(FakeGateDriver, self).__init__(*args, **kwargs)

    def _update_volume_stats(self):
        super(FakeGateDriver, self)._update_volume_stats()
        self._stats["pools"][0]["consistencygroup_support"] = True
        self._stats["pools"][0]["replication_enabled"] = True

    # NOTE(xyang): Consistency Group functions implemented below
    # are for testing purpose only. Data consistency cannot be
    # achieved by running these functions.
    def create_consistencygroup(self, context, group):
        """Creates a consistencygroup."""
        # A consistencygroup entry is already created in db
        # This driver just returns a status
        now = timeutils.utcnow()
        model_update = {'status': fields.ConsistencyGroupStatus.AVAILABLE,
                        'updated_at': now}

        return model_update

    def create_consistencygroup_from_src(self, context, group, volumes,
                                         cgsnapshot=None, snapshots=None,
                                         soure_cg=None, source_vols=None):
        """Creates a consistencygroup from cgsnapshot or source cg."""
        for vol in volumes:
            try:
                if snapshots:
                    for snapshot in snapshots:
                        if vol['snapshot_id'] == snapshot['id']:
                            self.create_volume_from_snapshot(vol, snapshot)
                            break
            except Exception:
                raise
            try:
                if source_vols:
                    for source_vol in source_vols:
                        if vol['source_volid'] == source_vol['id']:
                            self.create_cloned_volume(vol, source_vol)
                            break
            except Exception:
                raise
        return None, None

    def delete_consistencygroup(self, context, group, volumes):
        """Deletes a consistencygroup and volumes in the group."""
        model_update = {'status': group.status}
        volume_model_updates = []
        for volume_ref in volumes:
            volume_model_update = {'id': volume_ref.id}
            try:
                self.remove_export(context, volume_ref)
                self.delete_volume(volume_ref)
                volume_model_update['status'] = 'deleted'
            except exception.VolumeIsBusy:
                volume_model_update['status'] = 'available'
            except Exception:
                volume_model_update['status'] = 'error'
                model_update['status'] = fields.ConsistencyGroupStatus.ERROR
            volume_model_updates.append(volume_model_update)

        return model_update, volume_model_updates

    def update_consistencygroup(self, context, group,
                                add_volumes=None, remove_volumes=None):
        """Updates a consistency group."""
        return None, None, None

    def create_cgsnapshot(self, context, cgsnapshot, snapshots):
        """Creates a cgsnapshot.

        Snapshots created here are NOT consistent. This is for
        testing purpose only.
        """
        model_update = {'status': 'available'}
        snapshot_model_updates = []
        for snapshot in snapshots:
            snapshot_model_update = {'id': snapshot.id}
            try:
                self.create_snapshot(snapshot)
                snapshot_model_update['status'] = 'available'
            except Exception:
                snapshot_model_update['status'] = 'error'
                model_update['status'] = 'error'
            snapshot_model_updates.append(snapshot_model_update)

        return model_update, snapshot_model_updates

    def delete_cgsnapshot(self, context, cgsnapshot, snapshots):
        """Deletes a cgsnapshot."""
        model_update = {'status': cgsnapshot.status}
        snapshot_model_updates = []
        for snapshot in snapshots:
            snapshot_model_update = {'id': snapshot.id}
            try:
                self.delete_snapshot(snapshot)
                snapshot_model_update['status'] = 'deleted'
            except exception.SnapshotIsBusy:
                snapshot_model_update['status'] = 'available'
            except Exception:
                snapshot_model_update['status'] = 'error'
                model_update['status'] = 'error'
            snapshot_model_updates.append(snapshot_model_update)

        return model_update, snapshot_model_updates

    # Replication functions here are not really doing replication.
    # They are added so that we can do basic sanity check of replication
    # APIs.
    def replication_enable(self, context, volume):
        return

    def replication_disable(self, context, volume):
        return

    def replication_failover(self, context, volume, secondary):
        return {'model_update': {'status': volume['status']},
                'replication_driver_data': {'replication_driver_data': ''}}

    def list_replication_targets(self, context, volume):
        targets = []
        remote_target = {'managed_backend_name': None,
                         'type': 'unmanaged',
                         'remote_device_id': 'fake_remote_device',
                         'san_ip': '123.456.78.90'}
        targets.append(remote_target)
        return {'volume_id': volume['id'],
                'targets': targets}



# Copyright (c) 2013 eBay Inc.
# Copyright (c) 2013 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""The QoS specs extension"""

from oslo_log import log as logging
from oslo_utils import strutils
import six
import webob

from cinder.api import common
from cinder.api import extensions
from cinder.api.openstack import wsgi
from cinder.api.views import qos_specs as view_qos_specs
from cinder.api import xmlutil
from cinder import exception
from cinder.i18n import _, _LI
from cinder import rpc
from cinder import utils
from cinder.volume import qos_specs


LOG = logging.getLogger(__name__)

authorize = extensions.extension_authorizer('volume', 'qos_specs_manage')


def make_qos_specs(elem):
    elem.set('id')
    elem.set('name')
    elem.set('consumer')
    elem.append(SpecsTemplate())


def make_associations(elem):
    elem.set('association_type')
    elem.set('name')
    elem.set('id')


class SpecsTemplate(xmlutil.TemplateBuilder):
    def construct(self):
        return xmlutil.MasterTemplate(xmlutil.make_flat_dict('specs'), 1)


class QoSSpecsTemplate(xmlutil.TemplateBuilder):
    def construct(self):
        root = xmlutil.TemplateElement('qos_specs')
        elem = xmlutil.SubTemplateElement(root, 'qos_spec',
                                          selector='qos_specs')
        make_qos_specs(elem)
        return xmlutil.MasterTemplate(root, 1)


class QoSSpecsKeyDeserializer(wsgi.XMLDeserializer):
    def _extract_keys(self, key_node):
        keys = []
        for key in key_node.childNodes:
            key_name = key.tagName
            keys.append(key_name)

        return keys

    def default(self, string):
        dom = utils.safe_minidom_parse_string(string)
        key_node = self.find_first_child_named(dom, 'keys')
        if not key_node:
            LOG.info(_LI("Unable to parse XML input."))
            msg = _("Unable to parse XML request. "
                    "Please provide XML in correct format.")
            raise webob.exc.HTTPBadRequest(explanation=msg)
        return {'body': {'keys': self._extract_keys(key_node)}}


class AssociationsTemplate(xmlutil.TemplateBuilder):
    def construct(self):
        root = xmlutil.TemplateElement('qos_associations')
        elem = xmlutil.SubTemplateElement(root, 'associations',
                                          selector='qos_associations')
        make_associations(elem)
        return xmlutil.MasterTemplate(root, 1)


def _check_specs(context, specs_id):
    try:
        qos_specs.get_qos_specs(context, specs_id)
    except exception.QoSSpecsNotFound as ex:
        raise webob.exc.HTTPNotFound(explanation=six.text_type(ex))


class QoSSpecsController(wsgi.Controller):
    """The volume type extra specs API controller for the OpenStack API."""

    _view_builder_class = view_qos_specs.ViewBuilder

    @staticmethod
    def _notify_qos_specs_error(context, method, payload):
        rpc.get_notifier('QoSSpecs').error(context,
                                           method,
                                           payload)

    @wsgi.serializers(xml=QoSSpecsTemplate)
    def index(self, req):
        """Returns the list of qos_specs."""
        context = req.environ['cinder.context']
        authorize(context)

        params = req.params.copy()

        marker, limit, offset = common.get_pagination_params(params)
        sort_keys, sort_dirs = common.get_sort_params(params)
        filters = params
        allowed_search_options = ('id', 'name', 'consumer')
        utils.remove_invalid_filter_options(context, filters,
                                            allowed_search_options)

        specs = qos_specs.get_all_specs(context, filters=filters,
                                        marker=marker, limit=limit,
                                        offset=offset, sort_keys=sort_keys,
                                        sort_dirs=sort_dirs)
        return self._view_builder.summary_list(req, specs)

    @wsgi.serializers(xml=QoSSpecsTemplate)
    def create(self, req, body=None):
        context = req.environ['cinder.context']
        authorize(context)

        self.assert_valid_body(body, 'qos_specs')

        specs = body['qos_specs']
        name = specs.get('name', None)
        if name is None:
            msg = _("Please specify a name for QoS specs.")
            raise webob.exc.HTTPBadRequest(explanation=msg)

        self.validate_string_length(name, 'name', min_length=1,
                                    max_length=255, remove_whitespaces=True)
        name = name.strip()

        try:
            qos_specs.create(context, name, specs)
            spec = qos_specs.get_qos_specs_by_name(context, name)
            notifier_info = dict(name=name, specs=specs)
            rpc.get_notifier('QoSSpecs').info(context,
                                              'qos_specs.create',
                                              notifier_info)
        except exception.InvalidQoSSpecs as err:
            notifier_err = dict(name=name, error_message=err)
            self._notify_qos_specs_error(context,
                                         'qos_specs.create',
                                         notifier_err)
            raise webob.exc.HTTPBadRequest(explanation=six.text_type(err))
        except exception.QoSSpecsExists as err:
            notifier_err = dict(name=name, error_message=err)
            self._notify_qos_specs_error(context,
                                         'qos_specs.create',
                                         notifier_err)
            raise webob.exc.HTTPConflict(explanation=six.text_type(err))
        except exception.QoSSpecsCreateFailed as err:
            notifier_err = dict(name=name, error_message=err)
            self._notify_qos_specs_error(context,
                                         'qos_specs.create',
                                         notifier_err)
            raise webob.exc.HTTPInternalServerError(
                explanation=six.text_type(err))

        return self._view_builder.detail(req, spec)

    @wsgi.serializers(xml=QoSSpecsTemplate)
    def update(self, req, id, body=None):
        context = req.environ['cinder.context']
        authorize(context)

        self.assert_valid_body(body, 'qos_specs')
        specs = body['qos_specs']
        try:
            qos_specs.update(context, id, specs)
            notifier_info = dict(id=id, specs=specs)
            rpc.get_notifier('QoSSpecs').info(context,
                                              'qos_specs.update',
                                              notifier_info)
        except exception.QoSSpecsNotFound as err:
            notifier_err = dict(id=id, error_message=err)
            self._notify_qos_specs_error(context,
                                         'qos_specs.update',
                                         notifier_err)
            raise webob.exc.HTTPNotFound(explanation=six.text_type(err))
        except exception.InvalidQoSSpecs as err:
            notifier_err = dict(id=id, error_message=err)
            self._notify_qos_specs_error(context,
                                         'qos_specs.update',
                                         notifier_err)
            raise webob.exc.HTTPBadRequest(explanation=six.text_type(err))
        except exception.QoSSpecsUpdateFailed as err:
            notifier_err = dict(id=id, error_message=err)
            self._notify_qos_specs_error(context,
                                         'qos_specs.update',
                                         notifier_err)
            raise webob.exc.HTTPInternalServerError(
                explanation=six.text_type(err))

        return body

    @wsgi.serializers(xml=QoSSpecsTemplate)
    def show(self, req, id):
        """Return a single qos spec item."""
        context = req.environ['cinder.context']
        authorize(context)

        try:
            spec = qos_specs.get_qos_specs(context, id)
        except exception.QoSSpecsNotFound as err:
            raise webob.exc.HTTPNotFound(explanation=six.text_type(err))

        return self._view_builder.detail(req, spec)

    def delete(self, req, id):
        """Deletes an existing qos specs."""
        context = req.environ['cinder.context']
        authorize(context)

        force = req.params.get('force', None)

        # Convert string to bool type in strict manner
        force = strutils.bool_from_string(force)
        LOG.debug("Delete qos_spec: %(id)s, force: %(force)s",
                  {'id': id, 'force': force})

        try:
            qos_specs.delete(context, id, force)
            notifier_info = dict(id=id)
            rpc.get_notifier('QoSSpecs').info(context,
                                              'qos_specs.delete',
                                              notifier_info)
        except exception.QoSSpecsNotFound as err:
            notifier_err = dict(id=id, error_message=err)
            self._notify_qos_specs_error(context,
                                         'qos_specs.delete',
                                         notifier_err)
            raise webob.exc.HTTPNotFound(explanation=six.text_type(err))
        except exception.QoSSpecsInUse as err:
            notifier_err = dict(id=id, error_message=err)
            self._notify_qos_specs_error(context,
                                         'qos_specs.delete',
                                         notifier_err)
            if force:
                msg = _('Failed to disassociate qos specs.')
                raise webob.exc.HTTPInternalServerError(explanation=msg)
            msg = _('Qos specs still in use.')
            raise webob.exc.HTTPBadRequest(explanation=msg)

        return webob.Response(status_int=202)

    @wsgi.deserializers(xml=QoSSpecsKeyDeserializer)
    def delete_keys(self, req, id, body):
        """Deletes specified keys in qos specs."""
        context = req.environ['cinder.context']
        authorize(context)

        if not (body and 'keys' in body
                and isinstance(body.get('keys'), list)):
            raise webob.exc.HTTPBadRequest()

        keys = body['keys']
        LOG.debug("Delete_key spec: %(id)s, keys: %(keys)s",
                  {'id': id, 'keys': keys})

        try:
            qos_specs.delete_keys(context, id, keys)
            notifier_info = dict(id=id)
            rpc.get_notifier('QoSSpecs').info(context, 'qos_specs.delete_keys',
                                              notifier_info)
        except exception.QoSSpecsNotFound as err:
            notifier_err = dict(id=id, error_message=err)
            self._notify_qos_specs_error(context,
                                         'qos_specs.delete_keys',
                                         notifier_err)
            raise webob.exc.HTTPNotFound(explanation=six.text_type(err))
        except exception.QoSSpecsKeyNotFound as err:
            notifier_err = dict(id=id, error_message=err)
            self._notify_qos_specs_error(context,
                                         'qos_specs.delete_keys',
                                         notifier_err)
            raise webob.exc.HTTPBadRequest(explanation=six.text_type(err))

        return webob.Response(status_int=202)

    @wsgi.serializers(xml=AssociationsTemplate)
    def associations(self, req, id):
        """List all associations of given qos specs."""
        context = req.environ['cinder.context']
        authorize(context)

        LOG.debug("Get associations for qos_spec id: %s", id)

        try:
            associates = qos_specs.get_associations(context, id)
            notifier_info = dict(id=id)
            rpc.get_notifier('QoSSpecs').info(context,
                                              'qos_specs.associations',
                                              notifier_info)
        except exception.QoSSpecsNotFound as err:
            notifier_err = dict(id=id, error_message=err)
            self._notify_qos_specs_error(context,
                                         'qos_specs.associations',
                                         notifier_err)
            raise webob.exc.HTTPNotFound(explanation=six.text_type(err))
        except exception.CinderException as err:
            notifier_err = dict(id=id, error_message=err)
            self._notify_qos_specs_error(context,
                                         'qos_specs.associations',
                                         notifier_err)
            raise webob.exc.HTTPInternalServerError(
                explanation=six.text_type(err))

        return self._view_builder.associations(req, associates)

    def associate(self, req, id):
        """Associate a qos specs with a volume type."""
        context = req.environ['cinder.context']
        authorize(context)

        type_id = req.params.get('vol_type_id', None)

        if not type_id:
            msg = _('Volume Type id must not be None.')
            notifier_err = dict(id=id, error_message=msg)
            self._notify_qos_specs_error(context,
                                         'qos_specs.delete',
                                         notifier_err)
            raise webob.exc.HTTPBadRequest(explanation=msg)
        LOG.debug("Associate qos_spec: %(id)s with type: %(type_id)s",
                  {'id': id, 'type_id': type_id})

        try:
            qos_specs.associate_qos_with_type(context, id, type_id)
            notifier_info = dict(id=id, type_id=type_id)
            rpc.get_notifier('QoSSpecs').info(context,
                                              'qos_specs.associate',
                                              notifier_info)
        except exception.VolumeTypeNotFound as err:
            notifier_err = dict(id=id, error_message=err)
            self._notify_qos_specs_error(context,
                                         'qos_specs.associate',
                                         notifier_err)
            raise webob.exc.HTTPNotFound(explanation=six.text_type(err))
        except exception.QoSSpecsNotFound as err:
            notifier_err = dict(id=id, error_message=err)
            self._notify_qos_specs_error(context,
                                         'qos_specs.associate',
                                         notifier_err)
            raise webob.exc.HTTPNotFound(explanation=six.text_type(err))
        except exception.InvalidVolumeType as err:
            notifier_err = dict(id=id, error_message=err)
            self._notify_qos_specs_error(context,
                                         'qos_specs.associate',
                                         notifier_err)
            self._notify_qos_specs_error(context,
                                         'qos_specs.associate',
                                         notifier_err)
            raise webob.exc.HTTPBadRequest(explanation=six.text_type(err))
        except exception.QoSSpecsAssociateFailed as err:
            notifier_err = dict(id=id, error_message=err)
            self._notify_qos_specs_error(context,
                                         'qos_specs.associate',
                                         notifier_err)
            raise webob.exc.HTTPInternalServerError(
                explanation=six.text_type(err))

        return webob.Response(status_int=202)

    def disassociate(self, req, id):
        """Disassociate a qos specs from a volume type."""
        context = req.environ['cinder.context']
        authorize(context)

        type_id = req.params.get('vol_type_id', None)

        if not type_id:
            msg = _('Volume Type id must not be None.')
            notifier_err = dict(id=id, error_message=msg)
            self._notify_qos_specs_error(context,
                                         'qos_specs.delete',
                                         notifier_err)
            raise webob.exc.HTTPBadRequest(explanation=msg)
        LOG.debug("Disassociate qos_spec: %(id)s from type: %(type_id)s",
                  {'id': id, 'type_id': type_id})

        try:
            qos_specs.disassociate_qos_specs(context, id, type_id)
            notifier_info = dict(id=id, type_id=type_id)
            rpc.get_notifier('QoSSpecs').info(context,
                                              'qos_specs.disassociate',
                                              notifier_info)
        except exception.VolumeTypeNotFound as err:
            notifier_err = dict(id=id, error_message=err)
            self._notify_qos_specs_error(context,
                                         'qos_specs.disassociate',
                                         notifier_err)
            raise webob.exc.HTTPNotFound(explanation=six.text_type(err))
        except exception.QoSSpecsNotFound as err:
            notifier_err = dict(id=id, error_message=err)
            self._notify_qos_specs_error(context,
                                         'qos_specs.disassociate',
                                         notifier_err)
            raise webob.exc.HTTPNotFound(explanation=six.text_type(err))
        except exception.QoSSpecsDisassociateFailed as err:
            notifier_err = dict(id=id, error_message=err)
            self._notify_qos_specs_error(context,
                                         'qos_specs.disassociate',
                                         notifier_err)
            raise webob.exc.HTTPInternalServerError(
                explanation=six.text_type(err))

        return webob.Response(status_int=202)

    def disassociate_all(self, req, id):
        """Disassociate a qos specs from all volume types."""
        context = req.environ['cinder.context']
        authorize(context)

        LOG.debug("Disassociate qos_spec: %s from all.", id)

        try:
            qos_specs.disassociate_all(context, id)
            notifier_info = dict(id=id)
            rpc.get_notifier('QoSSpecs').info(context,
                                              'qos_specs.disassociate_all',
                                              notifier_info)
        except exception.QoSSpecsNotFound as err:
            notifier_err = dict(id=id, error_message=err)
            self._notify_qos_specs_error(context,
                                         'qos_specs.disassociate_all',
                                         notifier_err)
            raise webob.exc.HTTPNotFound(explanation=six.text_type(err))
        except exception.QoSSpecsDisassociateFailed as err:
            notifier_err = dict(id=id, error_message=err)
            self._notify_qos_specs_error(context,
                                         'qos_specs.disassociate_all',
                                         notifier_err)
            raise webob.exc.HTTPInternalServerError(
                explanation=six.text_type(err))

        return webob.Response(status_int=202)


class Qos_specs_manage(extensions.ExtensionDescriptor):
    """QoS specs support."""

    name = "Qos_specs_manage"
    alias = "qos-specs"
    namespace = "http://docs.openstack.org/volume/ext/qos-specs/api/v1"
    updated = "2013-08-02T00:00:00+00:00"

    def get_resources(self):
        resources = []
        res = extensions.ResourceExtension(
            Qos_specs_manage.alias,
            QoSSpecsController(),
            member_actions={"associations": "GET",
                            "associate": "GET",
                            "disassociate": "GET",
                            "disassociate_all": "GET",
                            "delete_keys": "PUT"})

        resources.append(res)

        return resources
